import os
import sys
import traceback
from collections import defaultdict
from typing import BinaryIO, Sequence, Optional, List, Dict, Union, Callable

import pytest
from pytest import Item, Collector

try:
    from pytest import CollectReport
except ImportError:
    from _pytest.reports import CollectReport  # 兼容pytest低版本

from testsolar_testtool_sdk.model.load import LoadResult, LoadError
from testsolar_testtool_sdk.model.param import EntryParam
from testsolar_testtool_sdk.model.test import TestCase
from testsolar_testtool_sdk.reporter import Reporter

from .converter import selector_to_pytest, pytest_to_selector, CASE_DRIVE_SEPARATOR
from .filter import filter_invalid_selector_path
from .parser import parse_case_attributes


def collect_testcases(
    entry_param: EntryParam,
    pipe_io: Optional[BinaryIO] = None,
    case_comment_fields: Optional[List[str]] = None,
    extra_load_function: Optional[
        Callable[[str, LoadResult, Dict[str, List[str]]], None]
    ] = None,
) -> None:
    if entry_param.ProjectPath not in sys.path:
        sys.path.insert(0, entry_param.ProjectPath)

    load_result: LoadResult = LoadResult(
        Tests=[],
        LoadErrors=[],
    )

    valid_selectors, load_errors = filter_invalid_selector_path(
        workspace=entry_param.ProjectPath,
        selectors=entry_param.TestSelectors,
    )

    load_result.LoadErrors.extend(load_errors)

    case_drive_records: Dict[str, List[str]] = defaultdict(list)
    pytest_paths: List[str] = []
    for selector in valid_selectors:
        # 扫描用例是否是基础用例，如果是存入 case_drive_records，方便后续扩展
        if CASE_DRIVE_SEPARATOR in selector:
            case_name, _, drive_key = selector.partition(CASE_DRIVE_SEPARATOR)
            case_drive_records[case_name].append(drive_key)

            pytest_paths.append(selector_to_pytest(test_selector=case_name))
        else:
            pytest_paths.append(selector_to_pytest(test_selector=selector))

    testcase_list = [
        os.path.join(entry_param.ProjectPath, it) for it in pytest_paths if it
    ]

    my_plugin = PytestCollector(pipe_io)
    args = [
        f"--rootdir={entry_param.ProjectPath}",
        "--collect-only",
        "--continue-on-collection-errors",
        "-v",
    ]
    args.extend(testcase_list)

    print(f"[Load] try to collect testcases: {args}")
    exit_code = pytest.main(args, plugins=[my_plugin])

    if exit_code != 0:
        print(f"[Warn][Load] collect testcases exit_code: {exit_code}")

    for item in my_plugin.collected:
        full_name = pytest_to_selector(item, entry_param.ProjectPath)
        attributes = parse_case_attributes(item, case_comment_fields)
        load_result.Tests.append(TestCase(Name=full_name, Attributes=attributes))

    # 增加额外功能，方便外部接入
    if extra_load_function:
        extra_load_function(entry_param.ProjectPath, load_result, case_drive_records)
    load_result.Tests.sort(key=lambda x: x.Name)

    for k, v in my_plugin.errors.items():
        load_result.LoadErrors.append(
            LoadError(
                name=f"load error of selector: [{k}]",
                message=v,
            )
        )

    load_result.LoadErrors.sort(key=lambda x: x.name)

    print(f"[Load] collect testcase count: {len(load_result.Tests)}")
    print(f"[Load] collect load error count: {len(load_result.LoadErrors)}")

    reporter = Reporter(pipe_io=pipe_io)
    reporter.report_load_result(load_result)


class PytestCollector:
    def __init__(self, pipe_io: Optional[BinaryIO] = None):
        self.collected: List[Item] = []
        self.errors: Dict[str, str] = {}
        self.reporter: Reporter = Reporter(pipe_io=pipe_io)

    def pytest_collection_modifyitems(
        self, items: Sequence[Union[Item, Collector]]
    ) -> None:
        for item in items:
            if isinstance(item, Item):
                self.collected.append(item)

    def pytest_collectreport(self, report: CollectReport) -> None:
        if report.failed:
            path = report.fspath
            if path in self.errors:
                return
            path = os.path.splitext(path)[0].replace(os.path.sep, ".")
            try:
                __import__(path)
            except Exception as e:
                print(e)
                self.errors[report.fspath] = traceback.format_exc()



from datetime import datetime, timedelta
from typing import List

try:
    from pytest import TestReport
except ImportError:
    from _pytest.reports import TestReport  # 兼容pytest低版本
from testsolar_testtool_sdk.model.testresult import TestCaseLog, LogLevel


def gen_logs(report: TestReport) -> TestCaseLog:
    logs: List[str] = []
    if report.capstdout:
        logs.append(report.capstdout)
    if report.capstderr:
        logs.append(report.capstderr)
    if report.caplog:
        logs.append(report.caplog)

    log = "\n".join(logs)

    if report.failed:
        error_log = report.longreprtext
        if error_log:
            log += "\n\n"
            log += error_log

    return TestCaseLog(
        Time=datetime.now() - timedelta(report.duration),
        Level=LogLevel.ERROR if report.failed else LogLevel.INFO,
        Content=log,
    )



import json
import re
from typing import Dict, List, Optional
from pytest import Item


# 解析测试用例的属性字段
#
# 1. 从commit解析字段
# 支持解析注释中的额外属性
#
# 2. 从mark解析字段
# 支持 @pytest.mark.attributes({"key":"value"}) 这种用法
#
# 解析属性包括：
# - description
# - tag
# - owner
# - extra_attributes
def parse_case_attributes(
    item: Item, comment_fields: Optional[List[str]] = None
) -> Dict[str, str]:
    """parse testcase attributes"""
    desc: str = (str(item.function.__doc__) or "").strip()  # type: ignore
    attributes: Dict[str, str] = {"description": desc}
    if comment_fields:
        attributes.update(scan_comment_fields(desc, comment_fields))

    if not item.own_markers:
        return attributes
    for mark in item.own_markers:
        if not mark.args and mark.name != "attributes":
            attributes["tag"] = mark.name
        elif mark.args and mark.name == "owner":
            attributes["owner"] = str(mark.args[0])
        elif mark.name == "extra_attributes":
            extra_attr = {}
            attr_list = []
            for key in mark.args[0]:
                if mark.args[0][key] is None:
                    continue
                extra_attr[key] = mark.args[0][key]
                attr_list.append(extra_attr)
            attributes["extra_attributes"] = json.dumps(attr_list)
    return attributes


def handle_str_param(desc: str) -> Dict[str, str]:
    """handle string parameter

    解析注释中单行 a = b 或 a: b 为 (a, b)形式方便后续处理
    """
    results: Dict[str, str] = {}
    pattern = re.compile(r".*?(\w+)\s*[:=]\s*(.+)")
    for line in desc.splitlines():
        match = pattern.match(line)
        if match:
            key, value = match.groups()
            results[key.strip()] = value.strip()
    return results


def scan_comment_fields(desc: str, desc_fields: List[str]) -> Dict[str, str]:
    """
    从函数的注释中解析额外字段
    """
    all_fields = handle_str_param(desc)
    results: Dict[str, str] = {}
    for key, value in all_fields.items():
        if key not in desc_fields:
            continue
        if "," in value:
            mutil_value = [v.strip() for v in value.split(",")]
            results[key] = json.dumps(mutil_value)
        else:
            results[key] = value
    return results



import io
import unittest
from pathlib import Path
from typing import Dict, List

from testsolar_testtool_sdk.model.param import EntryParam
from testsolar_testtool_sdk.model.load import LoadResult
from testsolar_testtool_sdk.pipe_reader import read_load_result

from testsolar_pytestx.collector import collect_testcases


class CollectorTest(unittest.TestCase):
    testdata_dir: str = str(
        Path(__file__).parent.parent.absolute().joinpath("testdata")
    )

    def test_collect_testcases_when_selector_is_valid(self):
        entry = EntryParam(
            TaskId="aa",
            ProjectPath=self.testdata_dir,
            TestSelectors=[
                "test_normal_case.py?test_success",
                "aa/bb/cc/test_in_sub_class.py",
                "test_data_drive.py",
                "errors/test_import_error.py",
                "errors/test_load_error.py",
            ],
            FileReportPath="",
        )

        pipe_io = io.BytesIO()
        collect_testcases(entry, pipe_io)
        pipe_io.seek(0)

        re = read_load_result(pipe_io)

        self.assertEqual(len(re.Tests), 6)
        self.assertEqual(len(re.LoadErrors), 2)

        self.assertEqual(
            re.Tests[0].Name, "aa/bb/cc/test_in_sub_class.py?TestCompute/test_add"
        )
        self.assertEqual(re.Tests[1].Name, "test_data_drive.py?test_eval/[2+4-6]")
        self.assertEqual(re.Tests[2].Name, "test_data_drive.py?test_eval/[3+5-8]")
        self.assertEqual(re.Tests[3].Name, "test_data_drive.py?test_eval/[6*9-42]")
        self.assertEqual(
            re.Tests[4].Name,
            "test_data_drive.py?test_special_data_drive_name/[中文-分号+[id:32]]",
        )

        self.assertEqual(re.Tests[5].Name, "test_normal_case.py?test_success")
        self.assertEqual(re.Tests[5].Attributes["owner"], "foo")
        self.assertEqual(re.Tests[5].Attributes["description"], "测试获取答案")
        self.assertEqual(re.Tests[5].Attributes["tag"], "high")
        self.assertEqual(
            re.Tests[5].Attributes["extra_attributes"], '[{"env": ["AA", "BB"]}]'
        )

        self.assertEqual(
            re.LoadErrors[0].name,
            "load error of selector: [errors/test_import_error.py]",
        )
        self.assertIn(
            "ModuleNotFoundError: No module named 'bad_import'",
            re.LoadErrors[0].message,
        )
        self.assertEqual(
            re.LoadErrors[1].name, "load error of selector: [errors/test_load_error.py]"
        )
        self.assertIn("SyntaxError: ", re.LoadErrors[1].message)

    def test_collect_testcases_when_select_not_valid(self):
        entry = EntryParam(
            TaskId="aa",
            ProjectPath=self.testdata_dir,
            TestSelectors=[
                "test_data_drive.py",
                "test_not_exist.py",
            ],
            FileReportPath="",
        )

        pipe_io = io.BytesIO()
        collect_testcases(entry, pipe_io)
        pipe_io.seek(0)

        re = read_load_result(pipe_io)

        self.assertEqual(len(re.Tests), 4)
        self.assertEqual(len(re.LoadErrors), 1)
        self.assertIn(
            "test_not_exist.py does not exist, SKIP it", re.LoadErrors[0].message
        )

    def test_collect_testcases_with_utf8_chars(self):
        entry = EntryParam(
            TaskId="aa",
            ProjectPath=self.testdata_dir,
            TestSelectors=[
                "test_data_drive_zh_cn.py",
            ],
            FileReportPath="",
        )

        pipe_io = io.BytesIO()
        collect_testcases(entry, pipe_io)
        pipe_io.seek(0)

        re = read_load_result(pipe_io)

        self.assertEqual(len(re.Tests), 3)
        self.assertEqual(len(re.LoadErrors), 0)

        self.assertEqual(
            re.Tests[0].Name, "test_data_drive_zh_cn.py?test_include/[#?-#?^$%!/]"
        )
        self.assertEqual(
            re.Tests[1].Name, "test_data_drive_zh_cn.py?test_include/[中文-中文汉字]"
        )
        self.assertEqual(
            re.Tests[2].Name,
            "test_data_drive_zh_cn.py?test_include/[파일을 찾을 수 없습니다-ファイルが見つかりません]",
        )

    def test_collect_testcases_with_case_drive_separator(self):
        entry = EntryParam(
            TaskId="aa",
            ProjectPath=self.testdata_dir,
            TestSelectors=[
                "test_normal_case.py?test_success→压缩机测试",
                "test_normal_case.py?test_success→解压机测试",
                "test_normal_case.py?test_success→循环机测试",
            ],
            FileReportPath="",
        )

        case_records = {}

        def loader_extend(
            param_1: str, param_2: LoadResult, param_3: Dict[str, List[str]]
        ) -> None:
            case_records.update(param_3)

        pipe_io = io.BytesIO()
        collect_testcases(entry, pipe_io, extra_load_function=loader_extend)
        pipe_io.seek(0)

        re = read_load_result(pipe_io)

        self.assertEqual(len(re.Tests), 1)
        self.assertEqual(len(re.LoadErrors), 0)

        self.assertEqual(re.Tests[0].Name, "test_normal_case.py?test_success")

        self.assertEqual(len(case_records), 1)
        self.assertIn("test_normal_case.py?test_success", case_records)

        records = case_records["test_normal_case.py?test_success"]
        self.assertEqual(len(records), 3)
        self.assertEqual(records[0], "压缩机测试")
        self.assertEqual(records[1], "解压机测试")
        self.assertEqual(records[2], "循环机测试")



import re
import os
from typing import Tuple, Optional

from pytest import Item
from loguru import logger

CASE_DRIVE_SEPARATOR = "→"


def selector_to_pytest(test_selector: str) -> str:
    """translate from test selector format to pytest format"""
    path, _, testcase = test_selector.partition("?")

    if not testcase:
        return path

    if "&" in testcase:
        testcase_attrs = testcase.split("&")
        for attr in testcase_attrs:
            if "name=" in attr:
                testcase = attr[5:]
                break
            elif "=" not in attr:
                testcase = attr
                break
    else:
        if testcase.startswith("name="):
            testcase = testcase[5:]

    case, datadrive = extract_case_and_datadrive(testcase)

    if datadrive:
        datadrive = encode_datadrive(datadrive)

    case = case.replace("/", "::")
    # 数据驱动里面的/不用替换为::
    result = f"{path}::{case}"
    if datadrive:
        result += datadrive

    return result


def extract_case_and_datadrive(case_selector: str) -> Tuple[str, str]:
    """
    Extract case and datadrive from test case selector

    从用例名称中拆分用例和数据驱动名称，pytest的数据驱动为最终的/[....]，如果不存在则返回空即可
    """
    if case_selector.endswith("]"):
        if case_selector.count("/[") > 1:
            logger.warning(
                f"Selector {case_selector} has more than 1 `/[` .Please fix your case drive data."
            )
        else:
            # 以]结尾，并且前面有/[，那么确实是一个数据驱动
            # 例子： testa/testb.py?case_name/[data/myf9:y678]
            case, _, drive_data = case_selector.partition("/[")
            if drive_data:
                return case, f"[{drive_data}"
    elif CASE_DRIVE_SEPARATOR in case_selector:
        # 数据驱动在用例名称里面
        # 例子： testa/testb.py?case_name→data
        case, _, drive_key = case_selector.partition(CASE_DRIVE_SEPARATOR)
        return case, ""

    return case_selector, ""


def pytest_to_selector(item: Item, project_path: str) -> str:
    """
    translate from pytest format to test selector format
    """

    if hasattr(item, "path") and hasattr(item, "cls") and item.path:
        rel_path = os.path.relpath(item.path, project_path)
        name = item.name
        if item.cls:
            name = item.cls.__name__ + "/" + name
        name = decode_datadrive(name)
        full_name = f"{rel_path}?{name}"
    elif hasattr(item, "nodeid") and item.nodeid:
        full_name = normalize_testcase_name(item.nodeid)
    else:
        rel_path, _, name = item.location
        name = name.replace(".", "/")
        name = decode_datadrive(name)
        full_name = f"{rel_path}?{name}"

    return full_name


def encode_datadrive(name: str) -> str:
    if name.endswith("]") and "[" in name:
        name = name.encode("unicode_escape").decode()
        name = name.replace("/[", "[")
    return name


def decode_datadrive(name: str) -> str:
    """
    将数据驱动转换为utf8字符，对用户来说可读性更好。

    原因：pytest by default escapes any non-ascii characters used in unicode strings for the parametrization because it has several downsides.

    https://docs.pytest.org/en/7.0.x/how-to/parametrize.html

    test_include[\u4e2d\u6587-\u4e2d\u6587\u6c49\u5b57] -> test_include[中文-中文汉字]

    用例名称中不允许出现[，因此如果有，一定是数据驱动的开头
    """
    if name.endswith("]"):
        start_index = name.find("[")
        if start_index != -1:
            name = name.replace(name[start_index], "/[", 1)

        if re.search(r"\\u\w{4}", name):
            name = name.encode().decode("unicode_escape")

    return name


def normalize_testcase_name(name: str, sub_case_key: Optional[str] = None) -> str:
    """test_directory/test_module.py::TestExampleClass::test_example_function[datedrive]
    -> test_directory/test_module.py?TestExampleClass/test_example_function/[datedrive]
    """
    assert "::" in name
    name = name.replace("::", "?", 1).replace(  # 第一个分割符是文件，因此替换为?
        "::", "/"
    )  # 后续的分割符是测试用例名称，替换为/
    name = decode_datadrive(name)
    if sub_case_key:
        name += CASE_DRIVE_SEPARATOR + sub_case_key
    return name



import pathlib
from typing import Tuple, List

from testsolar_testtool_sdk.model.load import LoadError


def filter_invalid_selector_path(
    workspace: str, selectors: List[str]
) -> Tuple[List[str], List[LoadError]]:
    valid_selectors = []
    invalid_selectors = []
    for selector in selectors:
        path, _, _ = selector.partition("?")

        full_path = pathlib.Path(workspace, path).resolve()
        if not full_path.exists():
            message = f"[WARNING]Path {full_path} does not exist, SKIP it"
            print(message)
            invalid_selectors.append(
                LoadError(name=f"invalid selector [{selector}]", message=message)
            )
        else:
            valid_selectors.append(selector)

    return valid_selectors, invalid_selectors



import json
import os
import shutil
from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Dict, Optional

from dacite import from_dict
from testsolar_testtool_sdk.model.testresult import (
    TestCaseLog,
    LogLevel,
    TestResult,
    ResultType,
    TestCaseStep,
)


@dataclass
class StatusDetails:
    message: Optional[str] = None
    trace: Optional[str] = None


@dataclass
class Parameter:
    name: str
    value: str


@dataclass
class Step:
    name: str
    status: str
    start: int
    stop: int
    parameters: Optional[List[Parameter]] = field(default_factory=list)  # type: ignore
    steps: Optional[List["Step"]] = field(  # type: ignore
        default_factory=list
    )  # Note the forward reference for recursive type
    statusDetails: Optional[StatusDetails] = None


@dataclass
class AllureData:
    name: str
    status: str
    start: int
    stop: int
    uuid: str
    historyId: str
    testCaseId: str
    fullName: str
    steps: List[Step] = field(default_factory=list)
    labels: List[Dict[str, str]] = field(default_factory=list)


def check_allure_enable() -> bool:
    return os.getenv("TESTSOLAR_TTP_ENABLEALLURE", "") != ""


def initialization_allure_dir(allure_dir: str) -> None:
    """
    初始化 Allure 报告目录。
    如果指定的目录存在，则删除该目录及其所有内容。然后重新创建一个空目录。
    """
    # 检查目录是否存在
    if os.path.isdir(allure_dir):
        # 目录存在，删除目录及其所有内容
        shutil.rmtree(allure_dir)
    # 创建一个新的空目录
    os.makedirs(allure_dir, exist_ok=True)


def generate_allure_results(test_data: Dict[str, TestResult], file_name: str) -> None:
    print("Start to generate allure results")
    with open(file_name) as fp:
        allure_data = from_dict(data_class=AllureData, data=json.loads(fp.read()))
        full_name = allure_data.fullName.replace("#", ".")
        for testcase_name in test_data.keys():
            testcase_format_name = ".".join(
                testcase_name.replace(".py?", os.sep).split(os.sep)
            )
            if full_name != testcase_format_name:
                continue
            if allure_data.steps:
                step_info = gen_allure_step_info(allure_data.steps)
            test_data[testcase_name].Steps.clear()
            test_data[testcase_name].Steps.extend(step_info)


def format_allure_time(timestamp: float) -> datetime:
    return datetime.fromtimestamp(timestamp / 1000)


def gen_allure_step_info(steps: List[Step], index: int = 0) -> List[TestCaseStep]:
    print("Gen allure step")
    case_steps = []
    for step in steps:
        index += 1
        result = step.status
        result_type: ResultType
        if result == "passed":
            result_type = ResultType.SUCCEED
        elif result == "skipped":
            result_type = ResultType.IGNORED
        else:
            result_type = ResultType.FAILED

        log = "\n"
        if step.parameters:
            for param in step.parameters:
                log += "%-30s%-20s\n" % (
                    "key: {}".format(param.name),
                    "value: {}".format(param.value),
                )
        if step.statusDetails:
            if step.statusDetails.message and step.statusDetails.trace:
                log += step.statusDetails.message + step.statusDetails.trace
        log_info = TestCaseLog(
            Time=format_allure_time(step.start),
            Level=LogLevel.ERROR if result == "failed" else LogLevel.INFO,
            Content=log,
        )
        step_info = TestCaseStep(
            Title="{}: {}".format(".".join(list(str(index))), step.name),
            Logs=[log_info],
            StartTime=format_allure_time(step.start),
            EndTime=format_allure_time(step.stop),
            ResultType=result_type,
        )

        print("Get allure step from json file: ", step_info)
        case_steps.append(step_info)
        if step.steps:
            case_steps.extend(gen_allure_step_info(step.steps, index * 10))
    return case_steps



from unittest import TestCase
from unittest.mock import MagicMock

from testsolar_pytestx.converter import (
    selector_to_pytest,
    pytest_to_selector,
    extract_case_and_datadrive,
)


class InnerClass:
    pass


class Test(TestCase):
    def test_selector_to_pytest_without_datadrive(self):
        re = selector_to_pytest(
            "/data/tests/tests/test_data_drive_zh_cn.py?aa/bb/test_include"
        )
        self.assertEqual(
            re, "/data/tests/tests/test_data_drive_zh_cn.py::aa::bb::test_include"
        )

    def test_selector_to_pytest_with_datadrive(self):
        re = selector_to_pytest(
            "/data/tests/tests/test_data_drive_zh_cn.py?aa/bb/test_include/[#?/-#?^$%!]"
        )
        self.assertEqual(
            re,
            "/data/tests/tests/test_data_drive_zh_cn.py::aa::bb::test_include[#?/-#?^$%!]",
        )

    def test_selector_to_pytest_with_utf8_string(self):
        re = selector_to_pytest(
            "/data/tests/tests/test_data_drive_zh_cn.py?aa/bb/test_include/[中文-中文汉字]"
        )
        self.assertEqual(
            re,
            "/data/tests/tests/test_data_drive_zh_cn.py::aa::bb::test_include[\\u4e2d\\u6587-\\u4e2d\\u6587\\u6c49\\u5b57]",
        )

        re = selector_to_pytest(
            "/data/tests/tests/test_data_drive_zh_cn.py?aa/bb/test_include/[中文-[中文]汉字]"
        )
        self.assertEqual(
            re,
            "/data/tests/tests/test_data_drive_zh_cn.py::aa::bb::test_include[\\u4e2d\\u6587-[\\u4e2d\\u6587]\\u6c49\\u5b57]",
        )

    def test_pytest_path_cls_to_selector(self):
        mock = MagicMock()
        mock.nodeid = None
        mock.path = "/data/tests/tests/test_data_drive_zh_cn.py"
        mock.name = "test_include[2-8]"
        mock.cls = InnerClass

        re = pytest_to_selector(mock, "/data/tests/")

        self.assertEqual(
            re, "tests/test_data_drive_zh_cn.py?InnerClass/test_include/[2-8]"
        )

    def test_pytest_node_id_to_selector_without_datadrive(self):
        mock = MagicMock()
        mock.nodeid = "tests/test_data_drive_zh_cn.py::test_include"
        mock.path = None
        mock.cls = None

        re = pytest_to_selector(mock, "/data/tests/")

        self.assertEqual(re, "tests/test_data_drive_zh_cn.py?test_include")

    def test_pytest_node_id_to_selector_with_datadrive(self):
        mock = MagicMock()
        mock.nodeid = (
            "/data/tests/tests/test_data_drive_zh_cn.py::test_include[#?-/[中文:203]]"
        )
        mock.path = None
        mock.cls = None

        re = pytest_to_selector(mock, "/data/tests/")

        self.assertEqual(
            re,
            "/data/tests/tests/test_data_drive_zh_cn.py?test_include/[#?-/[中文:203]]",
        )

    def test_pytest_location_to_selector(self):
        mock = MagicMock()
        mock.nodeid = None
        mock.path = None
        mock.cls = None
        mock.location = ("tests/test_data_drive_zh_cn.py", 22, "test_include[AA-BB]")

        re = pytest_to_selector(mock, "/data/tests/")

        self.assertEqual(re, "tests/test_data_drive_zh_cn.py?test_include/[AA-BB]")


class TestExtractCaseAndDataDrive:
    # 测试正常的数据驱动情况
    def test_extract_case_and_datadrive_with_datadrive(self):
        assert extract_case_and_datadrive("a/b/c/[data]") == ("a/b/c", "[data]")

    # 测试数据驱动包含特殊字符的情况
    def test_extract_case_and_datadrive_special_chars(self):
        assert extract_case_and_datadrive("a/b/c/[data→test]") == (
            "a/b/c",
            "[data→test]",
        )

    # 测试数据驱动在用例名称内部的情况
    def test_extract_case_and_datadrive_inside_name(self):
        assert extract_case_and_datadrive("a/b/c/data→[test]") == (
            "a/b/c/data→[test]",
            "",
        )
        assert extract_case_and_datadrive("a/b/c/data→/[test]") == (
            "a/b/c/data→",
            "[test]",
        )

    # 测试没有数据驱动的情况
    def test_extract_case_and_datadrive_no_datadrive(self):
        assert extract_case_and_datadrive("a/b/c") == ("a/b/c", "")

    # 测试只有用例名称的情况
    def test_extract_case_and_datadrive_only_case(self):
        assert extract_case_and_datadrive("case") == ("case", "")

    # 测试复杂路径的情况
    def test_extract_case_and_datadrive_complex_path(self):
        assert extract_case_and_datadrive("a/b/c/d/e/[data]") == ("a/b/c/d/e", "[data]")

    # 测试数据驱动在路径的最后一部分但不是有效数据驱动的情况
    def test_extract_case_and_datadrive_invalid_datadrive(self):
        assert extract_case_and_datadrive("a/b/c/d/e/[data") == ("a/b/c/d/e/[data", "")



import os
import sys
from datetime import datetime, timedelta
from typing import BinaryIO, Optional, Dict, Any, List, Callable

import pytest
from loguru import logger
from pytest import Item, Session

try:
    from pytest import TestReport
except ImportError:
    from _pytest.reports import TestReport

from testsolar_testtool_sdk.model.param import EntryParam
from testsolar_testtool_sdk.model.test import TestCase
from testsolar_testtool_sdk.model.testresult import TestResult, ResultType, TestCaseStep
from testsolar_testtool_sdk.reporter import Reporter
from enum import Enum

from .case_log import gen_logs
from .converter import selector_to_pytest, normalize_testcase_name
from .extend.allure_extend import (
    check_allure_enable,
    initialization_allure_dir,
    generate_allure_results,
)
from .filter import filter_invalid_selector_path
from .parser import parse_case_attributes


class RunMode(Enum):
    SINGLE = "single"
    BATCH = "batch"


def run_testcases(
    entry: EntryParam,
    pipe_io: Optional[BinaryIO] = None,
    case_comment_fields: Optional[List[str]] = None,
    run_mode: Optional[RunMode] = RunMode.BATCH,
    extra_run_function: Optional[Callable[[str, str, List[str]], str]] = None,
) -> None:
    if entry.ProjectPath not in sys.path:
        sys.path.insert(0, entry.ProjectPath)

    valid_selectors, _ = filter_invalid_selector_path(
        workspace=entry.ProjectPath,
        selectors=entry.TestSelectors,
    )

    args = [
        f"--rootdir={entry.ProjectPath}",
        "--continue-on-collection-errors",
        "-v",
    ]

    # check allure
    enable_allure = check_allure_enable()
    if enable_allure:
        print("Start allure test")
        allure_dir = os.path.join(entry.ProjectPath, "allure_results")
        args.append("--alluredir={}".format(allure_dir))
        initialization_allure_dir(allure_dir)

    extra_args = os.environ.get("TESTSOLAR_TTP_EXTRAARGS", "")
    if extra_args:
        args.extend(extra_args.split())

    reporter: Reporter = Reporter(pipe_io=pipe_io)

    if run_mode == RunMode.SINGLE:
        for it in valid_selectors:
            serial_args = args.copy()

            if extra_run_function is None:
                logger.error(
                    "[Error] Extra run function is not set, Please check extra_run_function"
                )
                return
            data_drive_key = extra_run_function(it, entry.ProjectPath, serial_args)
            logger.info(f"Pytest single run args: {serial_args}")
            my_plugin = PytestExecutor(
                reporter=reporter,
                comment_fields=case_comment_fields,
                data_drive_key=data_drive_key,
            )
            pytest.main(serial_args, plugins=[my_plugin])
    else:
        # 注意：传递给pytest中的用例必须在执行时能找到，否则pytest会报错
        # TODO: pytest执行出错时，将用例都设置为IGNORED，并设置错误原因
        args.extend(
            [
                os.path.join(entry.ProjectPath, selector_to_pytest(it))
                for it in valid_selectors
            ]
        )
        logger.info(f"Pytest run args: {args}")
        my_plugin = PytestExecutor(
            reporter=reporter, comment_fields=case_comment_fields
        )
        pytest.main(args, plugins=[my_plugin])

    logger.info("pytest process exit")


class PytestExecutor:
    def __init__(
        self,
        reporter: Reporter,
        comment_fields: Optional[List[str]] = None,
        data_drive_key: Optional[str] = None,
    ) -> None:
        self.reporter: Reporter = reporter
        self.testcase_count = 0
        self.testdata: Dict[str, TestResult] = {}
        self.skipped_testcase: Dict[str, str] = {}
        self.comment_fields = comment_fields
        self.data_drive_key = data_drive_key

    def pytest_runtest_logstart(self, nodeid: str, location: Any) -> None:
        """
        Called at the start of running the runtest protocol for a single item.
        """

        # 通知ResultHouse用例开始运行
        testcase_name = normalize_testcase_name(nodeid, self.data_drive_key)

        test_result = TestResult(
            Test=TestCase(Name=testcase_name),
            ResultType=ResultType.RUNNING,
            StartTime=datetime.now(),
            Message="",
        )

        self.testdata[testcase_name] = test_result

        logger.info(f"{nodeid} start")

        self.reporter.report_case_result(test_result)

    def pytest_runtest_setup(self, item: Item) -> None:
        """
        Called to perform the setup phase for a test item.
        """

        # 在Setup阶段将用例的属性解析出来并设置到Test中
        testcase_name = normalize_testcase_name(item.nodeid, self.data_drive_key)
        test_result = self.testdata[testcase_name]
        if test_result:
            test_result.Test.Attributes = parse_case_attributes(
                item, self.comment_fields
            )

    def pytest_runtest_logreport(self, report: TestReport) -> None:
        """
        Process the TestReport produced for each of the setup, call and teardown runtest phases of an item.
        """
        logger.info(f"{report.nodeid} log report")

        testcase_name = normalize_testcase_name(report.nodeid, self.data_drive_key)
        test_result = self.testdata[testcase_name]

        step_end_time = datetime.now()

        result_type: ResultType
        if report.failed:
            result_type = ResultType.FAILED
        elif report.skipped:
            result_type = ResultType.IGNORED
        else:
            result_type = ResultType.SUCCEED

        if report.when == "setup":
            test_result.Steps.append(
                TestCaseStep(
                    Title="Setup",
                    Logs=[gen_logs(report)],
                    StartTime=step_end_time - timedelta(report.duration),
                    EndTime=step_end_time,
                    ResultType=result_type,
                )
            )

            test_result.ResultType = result_type

            if report.skipped and isinstance(report.longrepr, tuple):
                file, line, reason = report.longrepr
                print(f"Skipped {file}:{line}: {reason}")
                test_result.Message = reason[:1000]

        elif report.when == "call":
            self.testcase_count += 1

            test_result.Steps.append(
                TestCaseStep(
                    Title="Run TestCase",
                    Logs=[gen_logs(report)],
                    StartTime=step_end_time - timedelta(report.duration),
                    EndTime=step_end_time,
                    ResultType=result_type,
                )
            )

            print(
                f"[{self.__class__.__name__}] Testcase {report.nodeid} run {report.outcome},"
                f" total {self.testcase_count} testcases complete"
            )

            if not test_result.Message and report.failed:
                # 避免错误信息过长，因此仅获取前面最多1000个字符
                test_result.Message = report.longreprtext[:1000]

            test_result.ResultType = result_type

        elif report.when == "teardown":
            test_result.Steps.append(
                TestCaseStep(
                    Title="Teardown",
                    Logs=[gen_logs(report)],
                    StartTime=step_end_time - timedelta(report.duration),
                    EndTime=step_end_time,
                    ResultType=result_type,
                )
            )
            if not test_result.is_final():
                test_result.ResultType = result_type

    def pytest_runtest_logfinish(self, nodeid: str, location: Any) -> None:
        """
        Called at the end of running the runtest protocol for a single item.
        """
        testcase_name = normalize_testcase_name(nodeid, self.data_drive_key)

        test_result = self.testdata[testcase_name]
        test_result.EndTime = datetime.now()

        # 检查是否allure报告，如果是在统一生成json文件后再上报
        enable_allure = check_allure_enable()
        if not enable_allure:
            self.reporter.report_case_result(test_result)

            # 上报完成后测试记录就没有用了，删除以节省内存
            self.testdata.pop(testcase_name, None)

    def pytest_sessionfinish(self, session: Session, exitstatus: int) -> None:
        """
        allure json报告在所有用例运行完才能生成, 故在运行用例结束后生成result并上报
        """
        enable_allure = check_allure_enable()
        if not enable_allure:
            return
        allure_dir = session.config.option.allure_report_dir
        for file_name in os.listdir(allure_dir):
            if not file_name.endswith("result.json"):
                continue
            generate_allure_results(self.testdata, os.path.join(allure_dir, file_name))
        for _, test_result in self.testdata.items():
            self.reporter.report_case_result(test_result)



import io
import os
from pathlib import Path
from unittest import TestCase

from testsolar_testtool_sdk.model.testresult import ResultType
from testsolar_testtool_sdk.pipe_reader import read_test_result

from run import run_testcases_from_args


class TestExecuteEntry(TestCase):
    testdata_dir: str = str(
        Path(__file__).parent.parent.absolute().joinpath("testdata")
    )

    def test_run_testcases_from_args(self):
        os.environ["TESTSOLAR_TTP_ENABLEALLURE"] = "1"
        pipe_io = io.BytesIO()
        run_testcases_from_args(
            args=[
                "run.py",
                Path.joinpath(Path(self.testdata_dir), "allure_entry.json"),
            ],
            workspace=self.testdata_dir,
            pipe_io=pipe_io,
        )

        # testcase running
        pipe_io.seek(0)
        start = read_test_result(pipe_io)
        self.assertEqual(start.ResultType, ResultType.RUNNING)
        self.assertEqual(start.Test.Name, "allure/allure_step_test.py?test_step")

        # testcase finish
        stop = read_test_result(pipe_io)
        self.assertEqual(stop.ResultType, ResultType.SUCCEED)
        self.assertEqual(stop.Test.Name, "allure/allure_step_test.py?test_step")
        self.assertEqual(stop.Message, "")
        self.assertEqual(type(stop.Steps), list)
        self.assertEqual(len(stop.Steps), 5)



import pytest
from unittest.mock import MagicMock
from testsolar_pytestx.parser import (
    parse_case_attributes,
    handle_str_param,
    scan_comment_fields,
)


def test_handle_str_param():
    desc = """
    description: This is a test function
    owner: test_owner
    tag: test_tag
    """
    expected = {
        "description": "This is a test function",
        "owner": "test_owner",
        "tag": "test_tag",
    }
    result = handle_str_param(desc)
    assert result == expected


def test_scan_comment_fields():
    desc = """
    description: This is a test function
    owner: test_owner
    tag: test_tag
    coding_testcase_id: 123,456
    """
    desc_fields = ["description", "owner", "coding_testcase_id"]
    expected = {
        "description": "This is a test function",
        "owner": "test_owner",
        "coding_testcase_id": '["123", "456"]',
    }
    assert scan_comment_fields(desc, desc_fields) == expected


def test_parse_case_attributes():
    # Mocking a pytest Item
    item = MagicMock()
    item.function.__doc__ = """
    description: This is a test function
    owner: test_owner
    tag: test_tag
    coding_testcase_id: 123,456
    """
    item.own_markers = []

    comment_fields = ["description", "owner", "coding_testcase_id"]
    expected = {
        "description": "This is a test function",
        "owner": "test_owner",
        "coding_testcase_id": '["123", "456"]',
    }
    assert parse_case_attributes(item, comment_fields) == expected


if __name__ == "__main__":
    pytest.main()



import io
from pathlib import Path
from unittest import TestCase

import dacite.exceptions
import pytest
from testsolar_testtool_sdk.model.testresult import ResultType
from testsolar_testtool_sdk.pipe_reader import read_test_result

from run import run_testcases_from_args


class TestExecuteEntry(TestCase):
    testdata_dir: str = str(
        Path(__file__).parent.parent.absolute().joinpath("testdata")
    )

    def test_run_testcases_from_args(self):
        pipe_io = io.BytesIO()
        run_testcases_from_args(
            args=[
                "run.py",
                Path.joinpath(Path(self.testdata_dir), "entry.json"),
            ],
            workspace=self.testdata_dir,
            pipe_io=pipe_io,
        )

        pipe_io.seek(0)
        start = read_test_result(pipe_io)
        self.assertEqual(start.ResultType, ResultType.RUNNING)

    def test_raise_error_when_param_is_invalid(self):
        with self.assertRaises(dacite.exceptions.MissingValueError):
            pipe_io = io.BytesIO()
            run_testcases_from_args(
                args=[
                    "run.py",
                    Path.joinpath(Path(self.testdata_dir), "bad_entry.json"),
                ],
                workspace=self.testdata_dir,
                pipe_io=pipe_io,
            )

    @pytest.mark.skip("暂时未实现，需要执行出错时上报忽略状态")
    def test_continue_run_when_one_case_is_not_found(self):
        pipe_io = io.BytesIO()
        run_testcases_from_args(
            args=[
                "run.py",
                Path.joinpath(Path(self.testdata_dir), "entry_1_case_not_found.json"),
            ],
            workspace=self.testdata_dir,
            pipe_io=pipe_io,
        )

        pipe_io.seek(0)
        start = read_test_result(pipe_io)
        self.assertEqual(start.ResultType, ResultType.IGNORED)



import io
from pathlib import Path
from unittest import TestCase

from testsolar_testtool_sdk.pipe_reader import read_load_result

from load import collect_testcases_from_args


class TestCollectorEntry(TestCase):
    testdata_dir: str = str(
        Path(__file__).parent.parent.absolute().joinpath("testdata")
    )

    def test_collect_testcases_from_args(self):
        pipe_io = io.BytesIO()
        collect_testcases_from_args(
            args=["load.py", Path.joinpath(Path(self.testdata_dir), "entry.json")],
            workspace=self.testdata_dir,
            pipe_io=pipe_io,
        )

        pipe_io.seek(0)
        re = read_load_result(pipe_io)

        self.assertEqual(len(re.Tests), 6)
        self.assertEqual(
            re.Tests[4].Name,
            "test_data_drive.py?test_special_data_drive_name/[中文-分号+[id:32]]",
        )

    def test_raise_error_when_param_is_invalid(self):
        with self.assertRaises(SystemExit):
            pipe_io = io.BytesIO()
            collect_testcases_from_args(
                args=["load.py"], workspace=self.testdata_dir, pipe_io=pipe_io
            )



import json
import sys
from pathlib import Path
from typing import Optional, List, BinaryIO

from dacite import from_dict
from testsolar_testtool_sdk.model.param import EntryParam

# 将pytestx加入path
parent = str(Path(__file__).parent.resolve())
if parent not in sys.path:
    sys.path.append(parent)

from testsolar_pytestx.executor import run_testcases  # noqa: E402


def run_testcases_from_args(
    args: List[str], workspace: Optional[str] = None, pipe_io: Optional[BinaryIO] = None
) -> None:
    if len(args) != 2:
        raise SystemExit("Usage: python run.py <entry_file>")

    filename = args[1]

    with open(filename, "r") as f:
        entry = from_dict(data_class=EntryParam, data=json.loads(f.read()))
        if workspace:
            entry.ProjectPath = workspace
        run_testcases(entry=entry, pipe_io=pipe_io)


if __name__ == "__main__":
    run_testcases_from_args(sys.argv)



import io
import unittest
from pathlib import Path

from testsolar_testtool_sdk.model.param import EntryParam
from testsolar_testtool_sdk.model.testresult import ResultType, LogLevel
from testsolar_testtool_sdk.pipe_reader import read_test_result

from testsolar_pytestx.executor import run_testcases


class ExecutorTest(unittest.TestCase):
    testdata_dir: str = str(
        Path(__file__).parent.parent.absolute().joinpath("testdata")
    )

    def test_run_success_testcase_with_logs(self):
        entry = EntryParam(
            TaskId="aa",
            ProjectPath=self.testdata_dir,
            TestSelectors=[
                "test_normal_case.py?name=test_success&tag=A&priority=High",
            ],
            FileReportPath="",
        )

        pipe_io = io.BytesIO()
        run_testcases(entry, pipe_io)
        pipe_io.seek(0)

        start = read_test_result(pipe_io)
        self.assertEqual(start.ResultType, ResultType.RUNNING)

        end = read_test_result(pipe_io)
        self.assertEqual(end.Test.Name, "test_normal_case.py?test_success")
        self.assertEqual(end.Test.Attributes["tag"], "high")
        self.assertEqual(end.Test.Attributes["owner"], "foo")
        self.assertEqual(end.ResultType, ResultType.SUCCEED)
        self.assertEqual(len(end.Steps), 3)

        step1 = end.Steps[0]
        self.assertEqual(step1.Title, "Setup")
        self.assertEqual(len(step1.Logs), 1)
        self.assertEqual(step1.Logs[0].Level, LogLevel.INFO)
        self.assertEqual(step1.ResultType, ResultType.SUCCEED)
        self.assertIn("this is setup", step1.Logs[0].Content)

        step2 = end.Steps[1]
        self.assertEqual(step2.Title, "Run TestCase")
        self.assertEqual(len(step2.Logs), 1)
        self.assertEqual(step2.Logs[0].Level, LogLevel.INFO)
        self.assertEqual(step2.ResultType, ResultType.SUCCEED)
        self.assertIn("this is print sample output", step2.Logs[0].Content)

        step3 = end.Steps[2]
        self.assertEqual(step3.Title, "Teardown")
        self.assertEqual(len(step3.Logs), 1)
        self.assertEqual(step3.Logs[0].Level, LogLevel.INFO)
        self.assertEqual(step3.ResultType, ResultType.SUCCEED)
        self.assertEqual(
            step3.Logs[0].Content,
            """this is setup
this is print sample output
this is teardown
""",
        )

    def test_run_success_testcase_with_one_invalid_selector(self):
        entry = EntryParam(
            TaskId="aa",
            ProjectPath=self.testdata_dir,
            TestSelectors=[
                "test_normal_case.py?name=test_success",
                "test_invalid_case.py?test_success",
            ],
            FileReportPath="",
        )

        pipe_io = io.BytesIO()
        run_testcases(entry, pipe_io)
        pipe_io.seek(0)

        start = read_test_result(pipe_io)
        self.assertEqual(start.ResultType, ResultType.RUNNING)

    def test_run_failed_testcase_with_log(self):
        entry = EntryParam(
            TaskId="aa",
            ProjectPath=self.testdata_dir,
            TestSelectors=[
                "test_normal_case.py?test_failed&priority=High",
            ],
            FileReportPath="",
        )

        pipe_io = io.BytesIO()
        run_testcases(entry, pipe_io)
        pipe_io.seek(0)

        start = read_test_result(pipe_io)
        self.assertEqual(start.ResultType, ResultType.RUNNING)

        end = read_test_result(pipe_io)
        self.assertEqual(end.ResultType, ResultType.FAILED)
        self.assertEqual(len(end.Steps), 3)
        self.assertIn("testdata/test_normal_case.py", end.Message)

        step2 = end.Steps[1]
        self.assertEqual(len(step2.Logs), 1)
        self.assertEqual(step2.Logs[0].Level, LogLevel.ERROR)
        self.assertEqual(step2.ResultType, ResultType.FAILED)
        self.assertIn("E       assert 4 == 6", step2.Logs[0].Content)

    def test_run_failed_testcase_with_raise_error(self):
        entry = EntryParam(
            TaskId="aa",
            ProjectPath=self.testdata_dir,
            TestSelectors=[
                "test_normal_case.py?test_raise_error",
            ],
            FileReportPath="",
        )

        pipe_io = io.BytesIO()
        run_testcases(entry, pipe_io)
        pipe_io.seek(0)

        start = read_test_result(pipe_io)
        self.assertEqual(start.ResultType, ResultType.RUNNING)

        end = read_test_result(pipe_io)
        self.assertEqual(end.ResultType, ResultType.FAILED)
        self.assertEqual(len(end.Steps), 3)

        step2 = end.Steps[1]
        self.assertEqual(len(step2.Logs), 1)
        self.assertEqual(step2.Logs[0].Level, LogLevel.ERROR)
        self.assertEqual(step2.ResultType, ResultType.FAILED)
        self.assertIn(
            "E       RuntimeError: this is raise runtime error", step2.Logs[0].Content
        )

    def test_run_skipped_testcase(self):
        entry = EntryParam(
            TaskId="aa",
            ProjectPath=self.testdata_dir,
            TestSelectors=[
                "test_skipped.py?test_filtered",
            ],
            FileReportPath="",
        )

        pipe_io = io.BytesIO()
        run_testcases(entry, pipe_io)
        pipe_io.seek(0)

        start = read_test_result(pipe_io)
        self.assertEqual(start.ResultType, ResultType.RUNNING)

        end = read_test_result(pipe_io)
        self.assertEqual(end.ResultType, ResultType.IGNORED)
        self.assertEqual(len(end.Steps), 2)
        self.assertEqual(end.Message, "Skipped: no way of currently testing this")

    def test_run_datadrive_with_single_value(self):
        entry = EntryParam(
            TaskId="aa",
            ProjectPath=self.testdata_dir,
            TestSelectors=[
                "test_data_drive.py?test_eval/[2+4-6]",
            ],
            FileReportPath="",
        )

        pipe_io = io.BytesIO()
        run_testcases(entry, pipe_io)
        pipe_io.seek(0)

        start = read_test_result(pipe_io)
        self.assertEqual(start.ResultType, ResultType.RUNNING)

        end = read_test_result(pipe_io)
        self.assertEqual(end.ResultType, ResultType.SUCCEED)
        self.assertEqual(len(end.Steps), 3)

    def test_run_datadrive_with_utf8_str(self):
        entry = EntryParam(
            TaskId="aa",
            ProjectPath=self.testdata_dir,
            TestSelectors=[
                "test_data_drive_zh_cn.py?test_include/[中文-中文汉字]",
            ],
            FileReportPath="",
        )

        pipe_io = io.BytesIO()
        run_testcases(entry, pipe_io)
        pipe_io.seek(0)

        start = read_test_result(pipe_io)
        self.assertEqual(start.ResultType, ResultType.RUNNING)

        end = read_test_result(pipe_io)
        self.assertEqual(end.ResultType, ResultType.SUCCEED)
        self.assertEqual(len(end.Steps), 3)
